#include <omp.h>
#include <cstdio>
#include <cstdlib>
#include <iostream>

#include "init.h"
#include "mylib.h"
#include "wave.h"
#include "kernels.h"

void GNhessian(float *hc11,float *hc13,float *hc33,const float *d0,const float *c11,const float *c13,const float *c33,const float *dc11,const float *dc13,const float *dc33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;

 memset(hc11,0,nnx*nnz*sizeof(float));
 memset(hc13,0,nnx*nnz*sizeof(float));
 memset(hc33,0,nnx*nnz*sizeof(float));
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 fprintf(stderr,"Total # GPUs = %d\n",nGPUs);

 float **d=new float*[nGPUs]();
 float **dd=new float*[nGPUs]();
 int **d_rloc=new int*[nGPUs]();
 float **d_c11=new float*[nGPUs]();
 float **d_c13=new float*[nGPUs]();
 float **d_c33=new float*[nGPUs]();
 float **d_dc11=new float*[nGPUs]();
 float **d_dc13=new float*[nGPUs]();
 float **d_dc33=new float*[nGPUs]();
 float **d_taper=new float*[nGPUs]();
 float **d_d=new float*[nGPUs]();
 float **p0=new float*[nGPUs]();
 float **q0=new float*[nGPUs]();
 float **p1=new float*[nGPUs]();
 float **q1=new float*[nGPUs]();
 float **tp=new float*[nGPUs]();
 float **tq=new float*[nGPUs]();
 float **Dp=new float*[nGPUs]();
 float **Dq=new float*[nGPUs]();
 float **d_hc11=new float*[nGPUs]();
 float **d_hc13=new float*[nGPUs]();
 float **d_hc33=new float*[nGPUs]();
 float **d_d0=new float*[nGPUs]();
 float **d_d1=new float*[nGPUs]();
 float **d_d00=new float*[nGPUs]();
 float **d_d01=new float*[nGPUs]();
 float **d_Dp0=new float*[nGPUs]();
 float **d_Dp1=new float*[nGPUs]();
 float **d_Dq0=new float*[nGPUs]();
 float **d_Dq1=new float*[nGPUs]();
 float **thc11=new float*[nGPUs]();
 float **thc13=new float*[nGPUs]();
 float **thc33=new float*[nGPUs]();
 
 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  
  init_coeffs();
 
  d[i]=new float[nr*nnt]();
  dd[i]=new float[nr*nnt]();
  
  cudaMalloc(&d_rloc[i],2*nr*sizeof(int));
  cudaMemcpy(d_rloc[i],rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
   
  cudaMalloc(&d_c11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_c11[i],c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c13[i],c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c33[i],c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_dc11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_dc13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_dc33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_dc11[i],dc11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_dc13[i],dc13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_dc33[i],dc33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_taper[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_taper[i],taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
  cudaMalloc(&d_d[i],nr*sizeof(float));
 
  cudaMalloc(&p0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&p1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&tp[i],nnx*nnz*sizeof(float)); cudaMemset(tp[i],0,nnx*nnz*sizeof(float));
  cudaMalloc(&tq[i],nnx*nnz*sizeof(float)); cudaMemset(tq[i],0,nnx*nnz*sizeof(float));
  
  Dp[i]=new float[nnx*nnz*nnt]();
  Dq[i]=new float[nnx*nnz*nnt]();
  
  cudaMalloc(&d_hc11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_hc13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_hc33[i],nnx*nnz*sizeof(float));
  
  cudaMalloc(&d_d0[i],nr*sizeof(float));
  cudaMalloc(&d_d1[i],nr*sizeof(float));
  cudaMalloc(&d_d00[i],nr*sizeof(float));
  cudaMalloc(&d_d01[i],nr*sizeof(float));
 
  cudaMalloc(&d_Dp0[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dq0[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dp1[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_Dq1[i],nnx*nnz*sizeof(float));
  
  thc11[i]=new float[nnx*nnz]();
  thc13[i]=new float[nnx*nnz]();
  thc33[i]=new float[nnx*nnz]();
 }

 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  fprintf(stderr,"Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   cudaSetDevice(i);

   int is=pass*nGPUs+i;
   int slocxz=sloc[0+is*2]+sloc[1+is*2]*nnx;

   dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
   dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
 
   inject<<<1,1>>>(p1[i],q1[i],wavelet[0],slocxz,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
  
   D<<<grid,block>>>(tp[i],tq[i],p1[i],q1[i],dx2,dz2);
  
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(d[i]+nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    cudaMemcpy(Dp[i]+nnx*nnz,tp[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
    cudaMemcpy(Dq[i]+nnx*nnz,tq[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
   // fprintf(stderr,"Time step it=%d\n",it);
  
    forwardC<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],tp[i],tq[i],d_c11[i],d_c13[i],d_c33[i],dt2);
  
    inject<<<1,1>>>(p0[i],q0[i],wavelet[it-1],slocxz,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
    
    D<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],dx2,dz2);
  
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(d[i]+(it/ratio)*nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
     cudaMemcpy(Dp[i]+(it/ratio)*nnx*nnz,tp[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
     cudaMemcpy(Dq[i]+(it/ratio)*nnx*nnz,tq[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
 
   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
  
   cudaMemcpy(d_Dp0[i],Dp[i],nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq0[i],Dq[i],nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dp1[i],Dp[i]+nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq1[i],Dq[i]+nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   inject<<<grid,block>>>(p1[i],q1[i],d_dc11[i],d_dc13[i],d_dc33[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],0.,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
  
   D<<<grid,block>>>(tp[i],tq[i],p1[i],q1[i],dx2,dz2);
  
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(dd[i]+nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
    //fprintf(stderr,"Time step it=%d\n",it);
  
    forwardC<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],tp[i],tq[i],d_c11[i],d_c13[i],d_c33[i],dt2);

    float f=(it-1.)/ratio;
    int i1=f,i2=min(i1+1,nnt-1);
    if(it%ratio==1){
     cudaMemcpy(d_Dp0[i],Dp[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     cudaMemcpy(d_Dq0[i],Dq[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_Dp0[i]; 
     d_Dp0[i]=d_Dp1[i];
     d_Dp1[i]=pt;
     pt=d_Dq0[i];
     d_Dq0[i]=d_Dq1[i];
     d_Dq1[i]=pt;
    }
    f=f-i1;
    inject<<<grid,block>>>(p0[i],q0[i],d_dc11[i],d_dc13[i],d_dc33[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],f,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
    
    D<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],dx2,dz2);
  
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(dd[i]+(it/ratio)*nr,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
   
   cudaMemset(d_hc11[i],0,nnx*nnz*sizeof(float));
   cudaMemset(d_hc13[i],0,nnx*nnz*sizeof(float));
   cudaMemset(d_hc33[i],0,nnx*nnz*sizeof(float));

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));
   
   cudaMemcpy(d_d0[i],dd[i]+(nt-1)/ratio*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
   inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p0[i],q0[i],d_d0[i],d_d1[i],0.,d_rloc[i],nr,nnt,dt2);
  
   abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
 
   float f=(nt-2.)/ratio;
   int i1=f,i2=i1+1;
   cudaMemcpy(d_Dp0[i],Dp[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq0[i],Dq[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dp1[i],Dp[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   cudaMemcpy(d_Dq1[i],Dq[i]+i2*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
   f=f-i1;
   imaging<<<grid,block>>>(d_hc11[i],d_hc13[i],d_hc33[i],p0[i],q0[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],f);
  
   for(int it=nt-3;it>=0;--it){
    //fprintf(stderr,"Time step it=%d\n",it);
    
    C<<<grid,block>>>(tp[i],tq[i],p0[i],q0[i],d_c11[i],d_c13[i],d_c33[i]);
  
    backwardD<<<grid,block>>>(p1[i],q1[i],p0[i],q0[i],tp[i],tq[i],dx2,dz2,dt2);
  
    f=(it+1.)/ratio;
    i1=f;
    if((it+2)%ratio==0){
     cudaMemcpy(d_d1[i],dd[i]+i1*nr,nr*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_d0[i]; 
     d_d0[i]=d_d1[i];
     d_d1[i]=pt;
    }
    f=f-i1;
    inject<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(p1[i],q1[i],d_d0[i],d_d1[i],f,d_rloc[i],nr,nnx,dt2);
  
	abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
  
	f=(float)it/ratio;
    i1=f;
    if((it+1)%ratio==0){
     cudaMemcpy(d_Dp1[i],Dp[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     cudaMemcpy(d_Dq1[i],Dq[i]+i1*nnx*nnz,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
     float *pt=d_Dp0[i]; 
     d_Dp0[i]=d_Dp1[i];
     d_Dp1[i]=pt;
     pt=d_Dq0[i];
     d_Dq0[i]=d_Dq1[i];
     d_Dq1[i]=pt;
    }
    f=f-i1;
    imaging<<<grid,block>>>(d_hc11[i],d_hc13[i],d_hc33[i],p1[i],q1[i],d_Dp0[i],d_Dq0[i],d_Dp1[i],d_Dq1[i],f);
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
   
   cudaMemcpy(thc11[i],d_hc11[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   cudaMemcpy(thc13[i],d_hc13[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
   cudaMemcpy(thc33[i],d_hc33[i],nnx*nnz*sizeof(float),cudaMemcpyDeviceToHost);
  }
  
  for(int i=0;i<nGPUsNeed;++i){
   #pragma omp parallel for num_threads(omp_get_num_threads()) shared(i)
   for(size_t ixz=0;ixz<nnx*nnz;++ixz){
    hc11[ixz]+=thc11[i][ixz];
    hc13[ixz]+=thc13[i][ixz];
    hc33[ixz]+=thc33[i][ixz];
   }
  }
  
  shotLeft-=nGPUsNeed;
 }

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  delete []d[i];delete []dd[i];
  cudaFree(d_rloc[i]);
  cudaFree(d_c11[i]);cudaFree(d_c13[i]);cudaFree(d_c33[i]); 
  cudaFree(d_dc11[i]);cudaFree(d_dc13[i]);cudaFree(d_dc33[i]); 
  cudaFree(d_taper[i]);
  cudaFree(d_d[i]);
  cudaFree(p0[i]);cudaFree(p1[i]);cudaFree(q0[i]);cudaFree(q1[i]);cudaFree(tp[i]);cudaFree(tq[i]);
  delete []Dp[i];delete []Dq[i]; 
  cudaFree(d_hc11[i]);cudaFree(d_hc13[i]);cudaFree(d_hc33[i]);
  cudaFree(d_d0[i]);cudaFree(d_d1[i]);cudaFree(d_d00[i]);cudaFree(d_d01[i]);
  cudaFree(d_Dp0[i]);cudaFree(d_Dp1[i]);cudaFree(d_Dq0[i]);cudaFree(d_Dq1[i]);
  delete []thc11[i];delete []thc13[i];delete []thc33[i];
 }
 
 delete []d;delete []dd;
 delete []d_rloc;
 delete []d_c11;delete []d_c13;delete []d_c33;
 delete []d_dc11;delete []d_dc13;delete []d_dc33;
 delete []d_taper;
 delete []d_d;
 delete []p0;delete []p1;delete []q0;delete []q1;delete []tp;delete []tq;
 delete []Dp;delete []Dq;
 delete []d_hc11;delete []d_hc13;delete []d_hc33;
 delete []d_d0;delete []d_d1;delete []d_d00;delete []d_d01;
 delete []d_Dp0;delete []d_Dp1;delete []d_Dq0;delete []d_Dq1;
 delete []thc11;delete []thc13;delete []thc33;

 return;
}
