module abc_mod
 implicit none
 real,allocatable::taper(:)
 real,parameter::pi=4.*atan(1.)
 contains

 subroutine init_abc(npad)
  integer,intent(in)::npad
  integer::i
  allocate(taper(npad))
  !$omp parallel do default(shared) private(i)
  do i=1,npad
   taper(i)=0.95+0.05*cos(pi*(real(npad-i))/npad)
  enddo
  !$omp end parallel do
 end subroutine 
  
 subroutine abc2d(field,nx,nz,npad)
  real,pointer::field(:,:)
  integer,intent(in)::nx,nz,npad
  integer::i
  !$omp parallel do default(shared) private(i)
  do i=1,nx+2*npad
   field(i,1:npad)=field(i,1:npad)*taper
   field(i,nz+npad+1:nz+2*npad)=field(i,nz+npad+1:nz+2*npad)*taper(npad:1:-1)
  enddo
  !$omp end parallel do
 
  !$omp parallel do default(shared) private(i)
  do i=1,nz+2*npad
   field(1:npad,i)=field(1:npad,i)*taper
   field(nx+npad+1:nx+2*npad,i)=field(nx+npad+1:nx+2*npad,i)*taper(npad:1:-1)
  enddo
  !$omp end parallel do  
 end subroutine
 
 subroutine abc3d(field,nx,ny,nz,npad)
  real,pointer::field(:,:,:)
  integer,intent(in)::nx,ny,nz,npad
  integer :: ix,iy,iz
  !$omp parallel do default(shared) private(ix,iy)
  do iy=1,ny+2*npad
   do ix=1,nx+2*npad
    field(ix,iy,1:npad)=field(ix,iy,1:npad)*taper
    field(ix,iy,nz+npad+1:nz+2*npad)=field(ix,iy,nz+npad+1:nz+2*npad)*taper(npad:1:-1)
   enddo
  enddo
  !$omp end parallel do

  !$omp parallel do default(shared) private(ix,iz)
  do iz=1,nz+2*npad
   do ix=1,nx+2*npad
    field(ix,1:npad,iz)=field(ix,1:npad,iz)*taper
    field(ix,ny+npad+1:ny+2*npad,iz)=field(ix,ny+npad+1:ny+2*npad,iz)*taper(npad:1:-1)
   enddo
  enddo
  !$omp end parallel do

  !$omp parallel do default(shared) private(iz,iy)
  do iy=1,ny+2*npad
   do iz=1,nz+2*npad
    field(1:npad,iy,iz)=field(1:npad,iy,iz)*taper
    field(nx+npad+1:nx+2*npad,iy,iz)=field(nx+npad+1:nx+2*npad,iy,iz)*taper(npad:1:-1)
   enddo
  enddo
  !$omp end parallel do  
 end subroutine 
 
 subroutine clean_abc()
  deallocate(taper)
 end subroutine
  
end module
