!module to solve the acoustic vti nonlinear differential wave equation by rotated staggered fd for heterogeneous medium 

module acoustic_vti_modeling_mod1
 use abc_mod
 use ieee_arithmetic
 use ieee_exceptions
 use sep
 
 implicit none
 
 contains 
 
 subroutine acoustic_vti_modeling_kernel1(dat,shot_loc,rec_loc,v,eps,del,sou,nx,nz,nt,npad,dx,dz,dt,ratio)
  integer,intent(in)::nx,nz,nt,ratio,shot_loc(:,:,:),rec_loc(:,:,:),npad
  real,intent(in)::dx,dz,dt,v(:,:),eps(:,:),del(:,:),sou(:)
  real,intent(inout)::dat(:,:,:,:,:)
  real,allocatable,target::prev_p(:,:),cur_p(:,:),prev_q(:,:),cur_q(:,:) !p=sigma_x, q=q
  real,pointer::pt_prev_p(:,:),pt_cur_p(:,:),pt_prev_q(:,:),pt_cur_q(:,:),pt(:,:)  
  integer::it,ix,iz,ix_shot,iz_shot,ix_rec,iz_rec,nx_rec,nz_rec,nnx,nnz,i,j
  real,parameter::pi=4.*atan(1.),c0=-205./72.,c1=8./5.,c2=-1./5.,c3=8./315.,c4=-1./560.
  real::temp_p,temp_q
  
  nnx=nx+2*npad
  nnz=nz+2*npad
  allocate(cur_p(nnx,nnz),prev_p(nnx,nnz),cur_q(nnx,nnz),prev_q(nnx,nnz))

  nx_rec=size(rec_loc,1)
  nz_rec=size(rec_loc,2)
  
  call init_abc(npad)
  
  call OMP_SET_NUM_THREADS(16)
  
  do j=1,size(shot_loc,2)
   do i=1,size(shot_loc,1)
    ix_shot=shot_loc(i,j,1)
    iz_shot=shot_loc(i,j,2)
    
    prev_p=0.
    cur_p=0.
    pt_prev_p=>prev_p
    pt_cur_p=>cur_p
    
    prev_q=0.
    cur_q=0.
    pt_prev_q=>prev_q
    pt_cur_q=>cur_q
  
    do it=1,nt
     write(0,*) "ix_shot",ix_shot,"iz_shot",iz_shot,"time",it
     
     !$omp parallel do default(shared) private(ix,iz,temp_p,temp_q)
     do iz=5,nnz-4
      do ix=5,nnx-4
       temp_p=(c0*pt_cur_p(ix,iz)+&
               c1*(pt_cur_p(ix+1,iz)+pt_cur_p(ix-1,iz))+&
               c2*(pt_cur_p(ix+2,iz)+pt_cur_p(ix-2,iz))+&
               c3*(pt_cur_p(ix+3,iz)+pt_cur_p(ix-3,iz))+&
               c4*(pt_cur_p(ix+4,iz)+pt_cur_p(ix-4,iz)))/dx/dx
       temp_q=(c0*pt_cur_q(ix,iz)+&
               c1*(pt_cur_q(ix,iz+1)+pt_cur_q(ix,iz-1))+&
               c2*(pt_cur_q(ix,iz+2)+pt_cur_q(ix,iz-2))+&
               c3*(pt_cur_q(ix,iz+3)+pt_cur_q(ix,iz-3))+&
               c4*(pt_cur_q(ix,iz+4)+pt_cur_q(ix,iz-4)))/dz/dz                                
       pt_prev_p(ix,iz)=dt*dt*v(ix,iz)*v(ix,iz)*((1.+2.*eps(ix,iz))*temp_p+sqrt(1.+2.*del(ix,iz))*temp_q)+2.*pt_cur_p(ix,iz)-pt_prev_p(ix,iz)  
       pt_prev_q(ix,iz)=dt*dt*v(ix,iz)*v(ix,iz)*(sqrt(1.+2.*del(ix,iz))*temp_p+temp_q)+2.*pt_cur_q(ix,iz)-pt_prev_q(ix,iz)                                        
      enddo
     enddo
     !$omp end parallel do
     
     pt_prev_p(ix_shot,iz_shot)=pt_prev_p(ix_shot,iz_shot)+dt*dt*sou(it)
!     pt_prev_p(ix_shot,iz_shot+1)=pt_prev_p(ix_shot,iz_shot+1)-dt*dt*sou(it)
     pt_prev_q(ix_shot,iz_shot)=pt_prev_q(ix_shot,iz_shot)+dt*dt*sou(it)
!     pt_prev_q(ix_shot,iz_shot+1)=pt_prev_q(ix_shot,iz_shot+1)-dt*dt*sou(it)
     

     call abc2d(pt_prev_p,nx,nz,npad)
!     call abc2d(pt_cur_p,nx,nz,npad)
     call abc2d(pt_prev_q,nx,nz,npad)
!     call abc2d(pt_cur_q,nx,nz,npad)

     if (mod(it-1,ratio)==0) then
      !$omp parallel do default(shared) private(ix_rec,iz_rec)
      do iz_rec=1,nz_rec
       do ix_rec=1,nx_rec
        dat(i,j,ix_rec,iz_rec,(it-1)/ratio+1)=0.5*(pt_prev_p(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2))+pt_prev_q(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2)))

        !dat(i,j,ix_rec,iz_rec,(it-1)/ratio+1)=0.5*(pt_prev_p(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2))-&
        !                                           pt_prev_p(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2)+1)+&
        !                                           pt_prev_q(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2))-&
        !                                           pt_prev_q(rec_loc(ix_rec,iz_rec,1),rec_loc(ix_rec,iz_rec,2)+1))
       enddo
      enddo
      !$omp end parallel do
     endif

!     if (maxval(abs(cur)) > 1.e3) then
!      write(0,*) "blow up"
!      write(0,*) maxval(abs(cur))
!      exit  ! Exit the loop when the amplitude blows up
!     endif

     if (.not. ieee_is_finite(sum(cur_p))) then
      write(0,*) "nan"
      exit ! Exit the loop when NaN is found
     endif    
   
!     call abc2d(pt_prev_p,nx,nz,npad)
!     call abc2d(pt_cur_p,nx,nz,npad)
!     call abc2d(pt_prev_q,nx,nz,npad)
!     call abc2d(pt_cur_q,nx,nz,npad)
   
     pt=>pt_prev_p
     pt_prev_p=>pt_cur_p
     pt_cur_p=>pt 
     
     pt=>pt_prev_q
     pt_prev_q=>pt_cur_q
     pt_cur_q=>pt      
    enddo !time loop
   enddo !x_shot loop
  enddo  !z_shot loop
  
  deallocate(prev_p,cur_p,prev_q,cur_q)  
  nullify(pt_cur_p,pt_prev_p,pt_cur_q,pt_prev_q,pt) 
!  call clean_abc()
 end subroutine

end module
