#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cstring>
#include "mylib.h"
#include "myio.h"
#include "wave.h"

void check_stability(float *v,float *eps,float *del,float v0,float eps0,float del0,int nnx,int nnz,int npad){
 int count=0;
 #pragma omp parallel for reduction(+:count)
 for(size_t i=0;i<nnx*nnz;++i){
  if(del0*del[i]>eps0*eps[i]){
   del[i]=eps0*eps[i]/del0;
   count+=1;
  }
 }
 set(v,1500./v0,(npad+5)*nnx);
 memset(eps,0.,(npad+5)*nnx*sizeof(float));
 memset(del,0.,(npad+5)*nnx*sizeof(float));
 printf("Modified models at %d points\n",count);
 return;
}

void check_stability(float *c11,float *c13,float *c33,int nnx,int nnz,int npad){
 int count=0;
// float *temp=new float[nnx*nnz]();
 #pragma omp parallel for reduction(+:count)
 for(size_t i=0;i<nnx*nnz;++i){
  if(c13[i]*c13[i]>c11[i]*c33[i]){
   c13[i]=sqrt(fabs(c11[i]*c33[i]));
  // temp[i]=10.;
   count+=1;
  }
 }
// write("check",temp,nnx*nnz);
// to_header("check","n1",nnx,"o1",-10*npad,"d1",10);
// to_header("check","n2",nnz,"o2",-10*npad,"d2",10);
// delete []temp;
 set(c33,1500.*1500.,(npad+5)*nnx);
 memcpy(c11,c33,(npad+5)*nnx*sizeof(float));
 memcpy(c13,c33,(npad+5)*nnx*sizeof(float));
 printf("Modified models at %d points\n",count);
 return;
}

void check_stability(const float *c11,const float *c13,const float *c33,float *dc11,float *dc13,float *dc33,size_t n){
 int count=0;
 #pragma omp parallel for reduction(+:count)
 for(size_t i=0;i<n;++i){
  float tc11=c11[i]+dc11[i];
  float tc13=c13[i]+dc13[i];
  float tc33=c33[i]+dc33[i];
  if(tc13*tc13>tc11*tc33){
   tc13=sqrt(fabs(tc11*tc33));
   dc13[i]=tc13-c13[i];
   count+=1;
  }
 }
 printf("Modified perturbations at %d points\n",count);
 return;
}
