#include <cstdlib>
#include <cstring>
#include <omp.h>
#include <cmath>
#include "wave.h"

void Cij2VEpsDel(float *v,float *eps,float *del,const float *c11,const float *c13,const float *c33,size_t n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  eps[i]=(c11[i]/c33[i]-1.)*0.5; 
  del[i]=(c13[i]*c13[i]/c33[i]/c33[i]-1.)*0.5; 
  v[i]=sqrt(c33[i]);
 }
 return;
}

void VEpsDel2Cij(float *c11,float *c13,float *c33,const float *v,const float *eps,const float *del,float v0,float eps0,float del0,size_t n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  c33[i]=v0*v0*v[i]*v[i]; //c33=v*v
  c11[i]=c33[i]*(1.+2.*eps0*eps[i]); //c11=v*v*(1+2*eps)
  c13[i]=c33[i]*sqrt(1.+2.*del0*del[i]); //c13=v*v*sqrt(1+2*del)
 }
 return;
}

void GradCij2GradVEpsDel(float *gv,float *geps,float *gdel,const float *gc11,const float *gc13,const float *gc33,const float *v,const float *eps,const float *del,float v0,float eps0,float del0,size_t n){
 #pragma omp parallel for default(shared)
 for(int i=0;i<n;++i){
  gv[i]=2.*v0*v0*v[i]*(gc11[i]*(1.+2.*eps0*eps[i])+gc13[i]*sqrt(1.+2.*del0*del[i])+gc33[i]);
  geps[i]=2.*v0*v0*v[i]*v[i]*eps0*gc11[i];
//  gdel[i]=v0*v0*v[i]*v[i]*del0/sqrt(1.+2.*del0*del[i])*gc13[i];
 }
 memset(gdel,0,n*sizeof(float));
 return;
}
