module init2d_mod
 use sep
 use pad_mod

 implicit none

 contains 

 subroutine init_spaces(nx,nz,nt,dx,dz,dt,ox,oz,npad,nnx,nnz)
  integer,intent(inout)::nx,nz,nt,npad,nnx,nnz
  real,intent(inout)::dx,dz,dt,ox,oz
  call from_param("nx",nx)
  call from_param("nz",nz)
  call from_param("nt",nt)
  call from_param("dx",dx)
  call from_param("dz",dz)
  call from_param("dt",dt)
  call from_param("ox",ox)
  call from_param("oz",oz) 
  call from_param("npad",npad) 
  nnx=nx+2*npad
  nnz=nz+2*npad
 end subroutine
 
 subroutine init_sou(sou,f,dt)
  real,intent(inout)::sou(:),f
  real,intent(in)::dt
  real::t,tdelay
  integer::i
  real,parameter::pi=4.*atan(1.)
  call from_param("f",f)
  tdelay=1.2/f
  do i=1,size(sou)
   t=(i-1.)*dt
   sou(i)=(1.-2.*pi*pi*f*f*(t-tdelay)*(t-tdelay))*exp(-pi*pi*f*f*(t-tdelay)*(t-tdelay))
  enddo  
 end subroutine
 
 subroutine init_model(v,eps,del,nx,nz,npad)
  real,intent(inout)::v(:,:),eps(:,:),del(:,:)
  integer,intent(in)::nx,nz,npad
  call sep_read(v(npad+1:nx+npad,npad+1:nz+npad),"v")
  call sep_read(eps(npad+1:nx+npad,npad+1:nz+npad),"eps")
  call sep_read(del(npad+1:nx+npad,npad+1:nz+npad),"del")
  call pad2d(v,nx,nz,npad)
  call pad2d(eps,nx,nz,npad)
  call pad2d(del,nx,nz,npad)
 end subroutine
 
 subroutine init_perturb(dv,deps,ddel,nx,nz,npad)
  real,intent(inout)::dv(:,:),deps(:,:),ddel(:,:)
  integer,intent(in)::nx,nz,npad
  call sep_read(dv(npad+1:nx+npad,npad+1:nz+npad),"dv")
  call sep_read(deps(npad+1:nx+npad,npad+1:nz+npad),"deps")
  call sep_read(ddel(npad+1:nx+npad,npad+1:nz+npad),"ddel")
  call pad2d(dv,nx,nz,npad)
  call pad2d(deps,nx,nz,npad)
  call pad2d(ddel,nx,nz,npad)
 end subroutine
 
 subroutine init_shot(nx_shot,nz_shot,dx_shot,dz_shot,ox_shot,oz_shot)
  integer,intent(inout)::nx_shot,nz_shot
  real,intent(inout)::ox_shot,oz_shot,dx_shot,dz_shot
  call from_param("ox_shot",ox_shot) 
  call from_param("dx_shot",dx_shot)
  call from_param("nx_shot",nx_shot)
  call from_param("oz_shot",oz_shot)
  call from_param("dz_shot",dz_shot)
  call from_param("nz_shot",nz_shot)
 end subroutine 
 
 subroutine init_shot_loc(shot_loc,nx_shot,nz_shot,dx_shot,dz_shot,ox_shot,oz_shot,npad,dx,dz,ox,oz)
  integer::ix_shot,iz_shot
  integer,intent(inout)::shot_loc(:,:,:)
  integer,intent(in)::nx_shot,nz_shot,npad
  real,intent(in)::ox_shot,oz_shot,dx_shot,dz_shot,dx,dz,ox,oz
  !$omp parallel do default(shared) private(ix_shot,iz_shot)
  do iz_shot=1,nz_shot
   do ix_shot=1,nx_shot
    shot_loc(ix_shot,iz_shot,1)=(ox_shot+(ix_shot-1)*dx_shot-ox)/dx+1.5+npad
    shot_loc(ix_shot,iz_shot,2)=(oz_shot+(iz_shot-1)*dz_shot-oz)/dz+1.5+npad
   enddo
  enddo
  !$omp end parallel do
 end subroutine
 
 subroutine init_rec(nx_rec,nz_rec,dx_rec,dz_rec,ox_rec,oz_rec)
  integer,intent(inout)::nx_rec,nz_rec
  real,intent(inout)::ox_rec,oz_rec,dx_rec,dz_rec
  call from_param("ox_rec",ox_rec)
  call from_param("dx_rec",dx_rec)
  call from_param("nx_rec",nx_rec)
  call from_param("oz_rec",oz_rec)
  call from_param("dz_rec",dz_rec)
  call from_param("nz_rec",nz_rec)
 end subroutine 
 
 subroutine init_rec_loc(rec_loc,nx_rec,nz_rec,dx_rec,dz_rec,ox_rec,oz_rec,npad,dx,dz,ox,oz)
  integer::ix_rec,iz_rec
  integer,intent(inout)::rec_loc(:,:,:)
  integer,intent(in)::nx_rec,nz_rec,npad
  real,intent(in)::ox_rec,oz_rec,dx_rec,dz_rec,dx,dz,ox,oz
  !$omp parallel do default(shared) private(ix_rec,iz_rec)
  do iz_rec=1,nz_rec
   do ix_rec=1,nx_rec
    rec_loc(ix_rec,iz_rec,1)=(ox_rec+(ix_rec-1)*dx_rec-ox)/dx+1.5+npad
    rec_loc(ix_rec,iz_rec,2)=(oz_rec+(iz_rec-1)*dz_rec-oz)/dz+1.5+npad
   enddo
  enddo
  !$omp end parallel do
 end subroutine
 
end module
