#include <cstdio>
#include "init.h"
#include "kernels.h"

__global__ void injectD(float *tp,float *tq,const float *p,const float *q,float dx2,float dz2,float dt2){
 __shared__ float sp[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 __shared__ float sq[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 int six=threadIdx.x+RADIUS;
 int siz=threadIdx.y+RADIUS;

 sp[six][siz]=p[i];
 sq[six][siz]=q[i];
 
 if(threadIdx.x<RADIUS){
  sp[threadIdx.x][siz]=p[i-RADIUS];
  sp[six+blockDim.x][siz]=p[i+blockDim.x];
  sq[threadIdx.x][siz]=q[i-RADIUS];
  sq[six+blockDim.x][siz]=q[i+blockDim.x];
 }

 if(threadIdx.y<RADIUS){
  sp[six][threadIdx.y]=p[i-RADIUS*nnx];
  sp[six][siz+blockDim.y]=p[i+blockDim.y*nnx];
  sq[six][threadIdx.y]=q[i-RADIUS*nnx];
  sq[six][siz+blockDim.y]=q[i+blockDim.y*nnx];
 }

 __syncthreads();

 tp[i]+=dt2*(ca[0]*sp[six][siz]+ca[1]*(sp[six-1][siz]+sp[six+1][siz])
                               +ca[2]*(sp[six-2][siz]+sp[six+2][siz])
                               +ca[3]*(sp[six-3][siz]+sp[six+3][siz])
                               +ca[4]*(sp[six-4][siz]+sp[six+4][siz]))/dx2;
 tq[i]+=dt2*(ca[0]*sq[six][siz]+ca[1]*(sq[six][siz-1]+sq[six][siz+1])
                               +ca[2]*(sq[six][siz-2]+sq[six][siz+2])
                               +ca[3]*(sq[six][siz-3]+sq[six][siz+3])
                               +ca[4]*(sq[six][siz-4]+sq[six][siz+4]))/dz2;
 return;
}

__global__ void inject(float *p,float *q,const float *d00,const float *d01,const float *d0,const float *d1,float f,const int *rloc,int nr,int nnx,float dt2){
 int ir=threadIdx.x+blockIdx.x*blockDim.x;
 if(ir<nr){
  int i=rloc[0+ir*2]+rloc[1+ir*2]*nnx;
  float s=0.5*((1.-f)*(d0[ir]-d00[ir])+f*(d1[ir]-d01[ir]));
  p[i]+=dt2*s;
  q[i]+=dt2*s;
 }
 return;
}

__global__ void inject(float *p,float *q,const float *d0,const float *d1,float f,const int *rloc,int nr,int nnx,float dt2){
 int ir=threadIdx.x+blockIdx.x*blockDim.x;
 if(ir<nr){
  int i=rloc[0+ir*2]+rloc[1+ir*2]*nnx;
  float s=0.5*((1.-f)*d0[ir]+f*d1[ir]);
  p[i]+=dt2*s;
  q[i]+=dt2*s;
 }
 return;
}

__global__ void inject(float *p,float sval,int slocxz){
 p[slocxz]+=sval;
 return;
}

__global__ void record(float *d_p,const float *p,int nnx){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int i=ix+iz*nnx;
 d_p[i]=p[i];
 return;
}

__global__ void inject(float *p,float *q,float sval,int slocxz,float dt2){
 p[slocxz]+=dt2*sval;
 q[slocxz]+=dt2*sval;
 return;
}

__global__ void inject(float *p,float *q,const float *fp,const float *fq,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p[i]+=dt2*fp[i];
 q[i]+=dt2*fq[i];
 return;
}

__global__ void inject(float *p,float *q,const float *dc11,const float *dc13,const float *dc33,const float *Dp,const float *Dq,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p[i]+=dt2*(Dp[i]*dc11[i]+Dq[i]*dc13[i]);
 q[i]+=dt2*(Dp[i]*dc13[i]+Dq[i]*dc33[i]);
 return;
}

__global__ void inject(float *p,float *q,const float *dc11,const float *dc13,const float *dc33,const float *Dp0,const float *Dq0,const float *Dp1,const float *Dq1,float f,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 float Dp=(1.-f)*Dp0[i]+f*Dp1[i];
 float Dq=(1.-f)*Dq0[i]+f*Dq1[i];
 p[i]+=dt2*(Dp*dc11[i]+Dq*dc13[i]);
 q[i]+=dt2*(Dp*dc13[i]+Dq*dc33[i]);
 return;
}

__global__ void abc(float *p,float *q,const float *taper){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p[i]*=taper[i];
 q[i]*=taper[i];
 return;
}

__global__ void record(float *d,const float *p,const float *q,const int *rloc,int nr,int nnx){
 int ir=threadIdx.x+blockIdx.x*blockDim.x;
 if(ir<nr){
  int i=rloc[0+ir*2]+rloc[1+ir*2]*nnx;
  d[ir]=0.5*(p[i]+q[i]);
 }
 return;
}

__global__ void D(float *tp,float *tq,const float *p,const float *q,float dx2,float dz2){
 __shared__ float sp[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 __shared__ float sq[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 int six=threadIdx.x+RADIUS;
 int siz=threadIdx.y+RADIUS;

 sp[six][siz]=p[i];
 sq[six][siz]=q[i];
 
 if(threadIdx.x<RADIUS){
  sp[threadIdx.x][siz]=p[i-RADIUS];
  sp[six+blockDim.x][siz]=p[i+blockDim.x];
  sq[threadIdx.x][siz]=q[i-RADIUS];
  sq[six+blockDim.x][siz]=q[i+blockDim.x];
 }

 if(threadIdx.y<RADIUS){
  sp[six][threadIdx.y]=p[i-RADIUS*nnx];
  sp[six][siz+blockDim.y]=p[i+blockDim.y*nnx];
  sq[six][threadIdx.y]=q[i-RADIUS*nnx];
  sq[six][siz+blockDim.y]=q[i+blockDim.y*nnx];
 }

 __syncthreads();

 tp[i]=(ca[0]*sp[six][siz]+ca[1]*(sp[six-1][siz]+sp[six+1][siz])
                          +ca[2]*(sp[six-2][siz]+sp[six+2][siz])
                          +ca[3]*(sp[six-3][siz]+sp[six+3][siz])
                          +ca[4]*(sp[six-4][siz]+sp[six+4][siz]))/dx2;
 tq[i]=(ca[0]*sq[six][siz]+ca[1]*(sq[six][siz-1]+sq[six][siz+1])
                          +ca[2]*(sq[six][siz-2]+sq[six][siz+2])
                          +ca[3]*(sq[six][siz-3]+sq[six][siz+3])
                          +ca[4]*(sq[six][siz-4]+sq[six][siz+4]))/dz2;
 return;
}

__global__ void C(float *tpp,float *tqq,const float *tp,const float *tq,
                  const float *c11,const float *c13,const float *c33){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 tpp[i]=c11[i]*tp[i]+c13[i]*tq[i];
 tqq[i]=c13[i]*tp[i]+c33[i]*tq[i];
 return;
}

__global__ void dC(float *tpp,float *tqq,const float *tp0,const float *tq0,const float *tp1,const float *tq1,float f,const float *dc11,const float *dc13,const float *dc33){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 float tp=(1.-f)*tp0[i]+f*tp1[i];
 float tq=(1.-f)*tq0[i]+f*tq1[i];
 tpp[i]=dc11[i]*tp+dc13[i]*tq;
 tqq[i]=dc13[i]*tp+dc33[i]*tq;
 return;
}

__global__ void forward(float *p0,float *q0,float *p1,float *q1,float *tpp,float *tqq,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p0[i]=2.*p1[i]+dt2*tpp[i]-p0[i];
 q0[i]=2.*q1[i]+dt2*tqq[i]-q0[i];
 return;
}

__global__ void backward(float *p1,float *q1,float *p0,float *q0,float *tpp,float *tqq,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p1[i]=2.*p0[i]+dt2*tpp[i]-p1[i];
 q1[i]=2.*q0[i]+dt2*tqq[i]-q1[i];
 return;
}

__global__ void imaging(float *dc11,float *dc13,float *dc33,const float *ap,const float *aq,const float *Dp,const float *Dq){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 dc11[i]+=ap[i]*Dp[i];
 dc13[i]+=ap[i]*Dq[i]+aq[i]*Dp[i];
 dc33[i]+=aq[i]*Dq[i];
 return;
}

__global__ void imaging(float *dc11,float *dc13,float *dc33,const float *ap,const float *aq,const float *Dp0,const float *Dq0,const float *Dp1,const float *Dq1,float f){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 float Dp=(1.-f)*Dp0[i]+f*Dp1[i];
 float Dq=(1.-f)*Dq0[i]+f*Dq1[i];
 dc11[i]+=ap[i]*Dp;
 dc13[i]+=ap[i]*Dq+aq[i]*Dp;
 dc33[i]+=aq[i]*Dq;
 return;
}

__global__ void forwardCD(float *p0,float *q0,const float *p1,const float *q1,const float *c11,const float *c13,const float *c33,float dx2,float dz2,float dt2){
 __shared__ float sp[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 __shared__ float sq[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 int six=threadIdx.x+RADIUS;
 int siz=threadIdx.y+RADIUS;

 sp[six][siz]=p1[i];
 sq[six][siz]=q1[i];
 
 if(threadIdx.x<RADIUS){
  sp[threadIdx.x][siz]=p1[i-RADIUS];
  sp[six+blockDim.x][siz]=p1[i+blockDim.x];
  sq[threadIdx.x][siz]=q1[i-RADIUS];
  sq[six+blockDim.x][siz]=q1[i+blockDim.x];
 }

 if(threadIdx.y<RADIUS){
  sp[six][threadIdx.y]=p1[i-RADIUS*nnx];
  sp[six][siz+blockDim.y]=p1[i+blockDim.y*nnx];
  sq[six][threadIdx.y]=q1[i-RADIUS*nnx];
  sq[six][siz+blockDim.y]=q1[i+blockDim.y*nnx];
 }

 __syncthreads();

 float tp=(ca[0]*sp[six][siz]+ca[1]*(sp[six-1][siz]+sp[six+1][siz])
                             +ca[2]*(sp[six-2][siz]+sp[six+2][siz])
                             +ca[3]*(sp[six-3][siz]+sp[six+3][siz])
                             +ca[4]*(sp[six-4][siz]+sp[six+4][siz]))/dx2;
 float tq=(ca[0]*sq[six][siz]+ca[1]*(sq[six][siz-1]+sq[six][siz+1])
                             +ca[2]*(sq[six][siz-2]+sq[six][siz+2])
                             +ca[3]*(sq[six][siz-3]+sq[six][siz+3])
                             +ca[4]*(sq[six][siz-4]+sq[six][siz+4]))/dz2;
 p0[i]=2.*p1[i]+dt2*(c11[i]*tp+c13[i]*tq)-p0[i];
 q0[i]=2.*q1[i]+dt2*(c13[i]*tp+c33[i]*tq)-q0[i];
 return;
}

__global__ void backwardD(float *p1,float *q1,const float *p0,const float *q0,const float *tp,const float *tq,float dx2,float dz2,float dt2){
 __shared__ float sp[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 __shared__ float sq[BLOCK_DIM_X+2*RADIUS][BLOCK_DIM_Y+2*RADIUS]; 
 
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 int six=threadIdx.x+RADIUS;
 int siz=threadIdx.y+RADIUS;

 sp[six][siz]=tp[i];
 sq[six][siz]=tq[i];
 
 if(threadIdx.x<RADIUS){
  sp[threadIdx.x][siz]=tp[i-RADIUS];
  sp[six+blockDim.x][siz]=tp[i+blockDim.x];
  sq[threadIdx.x][siz]=tq[i-RADIUS];
  sq[six+blockDim.x][siz]=tq[i+blockDim.x];
 }

 if(threadIdx.y<RADIUS){
  sp[six][threadIdx.y]=tp[i-RADIUS*nnx];
  sp[six][siz+blockDim.y]=tp[i+blockDim.y*nnx];
  sq[six][threadIdx.y]=tq[i-RADIUS*nnx];
  sq[six][siz+blockDim.y]=tq[i+blockDim.y*nnx];
 }

 __syncthreads();

 float tp0=(ca[0]*sp[six][siz]+ca[1]*(sp[six-1][siz]+sp[six+1][siz])
                              +ca[2]*(sp[six-2][siz]+sp[six+2][siz])
                              +ca[3]*(sp[six-3][siz]+sp[six+3][siz])
                              +ca[4]*(sp[six-4][siz]+sp[six+4][siz]))/dx2;
 float tq0=(ca[0]*sq[six][siz]+ca[1]*(sq[six][siz-1]+sq[six][siz+1])
                              +ca[2]*(sq[six][siz-2]+sq[six][siz+2])
                              +ca[3]*(sq[six][siz-3]+sq[six][siz+3])
                              +ca[4]*(sq[six][siz-4]+sq[six][siz+4]))/dz2;
 p1[i]=2.*p0[i]+dt2*tp0-p1[i];
 q1[i]=2.*q0[i]+dt2*tq0-q1[i];
 return;
}

__global__ void forwardC(float *p0,float *q0,const float *p1,const float *q1,const float *tp,const float *tq,
                  const float *c11,const float *c13,const float *c33,float dt2){
 int ix=threadIdx.x+blockIdx.x*blockDim.x+RADIUS;
 int iz=threadIdx.y+blockIdx.y*blockDim.y+RADIUS;
 int nnx=blockDim.x*gridDim.x+2*RADIUS;
 int i=ix+iz*nnx;
 p0[i]=2.*p1[i]+dt2*(c11[i]*tp[i]+c13[i]*tq[i])-p0[i];
 q0[i]=2.*q1[i]+dt2*(c13[i]*tp[i]+c33[i]*tq[i])-q0[i];
 return;
}

void init_coeffs(){
 cudaMemcpyToSymbol(ca,a,(RADIUS+1)*sizeof(float));
 return;
}

__global__ void check_ca(){
 printf("ca %f %f %f %f %f\n",ca[0],ca[1],ca[2],ca[3],ca[4]);
 return;
}
