//Following musa1 in sep150
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include "mylib.h"
#include "wave.h"
#include "solvers.h"

float line_search(float *c11,float *c13,float *c33,float &f,float *ggc11,float *ggc13,float *ggc33,float f0,const float *gc11,const float *gc13,const float *gc33,const float *pc11,const float *pc13,const float *pc33,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 float a;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float g0=dot_product(gc11,pc11,nnx*nnz)+dot_product(gc13,pc13,nnx*nnz)+dot_product(gc33,pc33,nnx*nnz);
 float aold=0.,anew=a0,fold=f0,fnew,gold=g0,gnew;

 float *tc11=new float[nnx*nnz]();
 float *tc13=new float[nnx*nnz]();
 float *tc33=new float[nnx*nnz]();
 
 for(int i=0;i<NITER_LS;++i){
  lin_comb(tc11,1.,c11,anew,pc11,nnx*nnz);
  lin_comb(tc13,1.,c13,anew,pc13,nnx*nnz);
  lin_comb(tc33,1.,c33,anew,pc33,nnx*nnz);
  check_stability(tc11,tc13,tc33,nnx,nnz,npad);

  fnew=objFuncGradient(ggc11,ggc13,ggc33,d0,tc11,tc13,tc33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  
  fprintf(stderr,"line search iteration %d, new step length %f, new obj func %f\n",i,anew,fnew);
  
  gnew=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(isnan(fnew) || isinf(fnew)) anew*=0.9;

  if(fnew>f0+c1*anew*g0 || (fnew>=fold && i>0)){
   a=zoom(c11,c13,c33,f,ggc11,ggc13,ggc33,aold,anew,fold,fnew,gold,gnew,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  if(fabs(gnew)<=-c2*g0){
   fprintf(stderr,"found step length after %d line searches\n",i);
   a=anew;f=fnew;
   memcpy(c11,tc11,nnx*nnz*sizeof(float));
   memcpy(c13,tc13,nnx*nnz*sizeof(float));
   memcpy(c33,tc33,nnx*nnz*sizeof(float));
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }
  
  if(gnew>=0.){
   a=zoom(c11,c13,c33,f,ggc11,ggc13,ggc33,anew,aold,fnew,fold,gnew,gold,f0,g0,pc11,pc13,pc33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
   delete []tc11;delete []tc13;delete []tc33;
   return a;
  }

  aold=anew;
  fold=fnew;
  gold=gnew;
  anew*=2.;
 }

 fprintf(stderr,"line search fails in %d tries\n",NITER_LS);

 a=anew;f=fnew;
 
 memcpy(c11,tc11,nnx*nnz*sizeof(float));
 memcpy(c13,tc13,nnx*nnz*sizeof(float));
 memcpy(c33,tc33,nnx*nnz*sizeof(float));

 delete []tc11;delete []tc13;delete []tc33;

 return a;
}

float zoom(float *c11,float *c13,float *c33,float &f,float *ggc11,float *ggc13,float *ggc33,float low,float high,float flow,float fhi,float glow,float ghi,float f0,float g0,const float *pc11,const float *pc13,const float *pc33,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 float atr;
 
 int nnx=nx+2*npad,nnz=nz+2*npad; 
 
 float *c11tr=new float[nnx*nnz]();
 float *c13tr=new float[nnx*nnz]();
 float *c33tr=new float[nnx*nnz]();
 
 float low2=high,flow2=fhi,glow2=ghi;

 float ftr,gtr;
 int i=0;
 while(low!=high && i<NITER_ZOOM){
  fprintf(stderr,"zoom iteration %d alow %f ahigh %f\n",i,low,high);
  ++i;
  
  float bisect=0.5*(low+high);
  
  float d1=glow+glow2-3.*(flow-flow2)/(low-low2);
  float d2=sign(low2-low)*sqrt(d1*d1-glow*glow2);
  float interp=low2-(low2-low)*(glow2+d2-d1)/(glow2-glow+2.*d2);
  
  int in_int=0;
  if(isfinite(interp)){
   if(low<=bisect){
    if(interp>=low && interp<=bisect) in_int=1;
   }
   else{
    if(interp>=bisect && interp<=low) in_int=1;
   }
  }
  
  if(in_int) atr=interp;
  else atr=bisect;

  lin_comb(c11tr,1.,c11,atr,pc11,nnx*nnz);
  lin_comb(c13tr,1.,c13,atr,pc13,nnx*nnz);
  lin_comb(c33tr,1.,c33,atr,pc33,nnx*nnz);
  check_stability(c11tr,c13tr,c33tr,nnx,nnz,npad);

  ftr=objFuncGradient(ggc11,ggc13,ggc33,d0,c11tr,c13tr,c33tr,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  
  fprintf(stderr,"trial step length %f, trial obj func %f\n",atr,ftr);
  
  gtr=dot_product(ggc11,pc11,nnx*nnz)+dot_product(ggc13,pc13,nnx*nnz)+dot_product(ggc33,pc33,nnx*nnz);
  
  if(ftr>f0+c1*atr*g0 || ftr>=flow){
   high=atr;fhi=ftr;ghi=gtr;
   low2=atr;flow2=ftr;glow2=gtr;
  }
  else{
   if(fabs(gtr)<=-c2*g0){
    fprintf(stderr,"found step length after %d steps of zoom\n",i);
    memcpy(c11,c11tr,nnx*nnz*sizeof(float));
    memcpy(c13,c13tr,nnx*nnz*sizeof(float));
    memcpy(c33,c33tr,nnx*nnz*sizeof(float));
    
    delete []c11tr;delete []c13tr;delete []c33tr;

    f=ftr;
    return atr;
   }
   
   if(gtr*(high-low)>=0.){
    high=low;
	fhi=flow;
	ghi=glow;
   }

   low=atr;flow=ftr;glow=gtr;
   low2=high;flow2=fhi;glow2=ghi;
  }
 }
 
 fprintf(stderr,"zoom fails in %d tries\n",i);
 
 memcpy(c11,c11tr,nnx*nnz*sizeof(float));
 memcpy(c13,c13tr,nnx*nnz*sizeof(float));
 memcpy(c33,c33tr,nnx*nnz*sizeof(float));
 
 delete []c11tr;delete []c13tr;delete []c33tr;

 f=ftr;
 return atr;
}
