//Following musa1 in sep150
#include <cstdio>
#include <cstdlib>
#include "mylib.h"
#include "myio.h"
#include "wave.h"
#include "solvers.h"

void nlcg(float *c11,float *c13,float *c33,float *f,float *g,int niter,float a0,const float *d0,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){ 
 int nnx=nx+2*npad,nnz=nz+2*npad;
 
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 
 f[0]=objFuncGradient(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 
 g[0]=dot_product(gc11,gc11,nnx*nnz)+dot_product(gc13,gc13,nnx*nnz)+dot_product(gc33,gc33,nnx*nnz);

 write("objfunc",f,1);
 to_header("objfunc","n1",niter,"o1",0.,"d1",1.);
 
 write("normg2",g,1);
 to_header("normg2","n1",niter,"o1",0.,"d1",1.);
 
 float *pc11=new float[nnx*nnz](); mynegate(pc11,gc11,nnx*nnz);
 float *pc13=new float[nnx*nnz](); mynegate(pc13,gc13,nnx*nnz);
 float *pc33=new float[nnx*nnz](); mynegate(pc33,gc33,nnx*nnz);
 
 write("ic11",c11,nnx*nnz);
 write("ic13",c13,nnx*nnz);
 write("ic33",c33,nnx*nnz);

// write("ic11",ic11,nnx*nnz*niter);
 to_header("ic11","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("ic11","n2",nnz,"o2",-dz*npad,"d2",dz);
 to_header("ic11","n3",niter,"o3",0.,"d3",1.);
// write("ic13",ic13,nnx*nnz*niter);
 to_header("ic13","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("ic13","n2",nnz,"o2",-dz*npad,"d2",dz);
 to_header("ic13","n3",niter,"o3",0.,"d3",1.);
// write("ic33",ic33,nnx*nnz*niter);
 to_header("ic33","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("ic33","n2",nnz,"o2",-dz*npad,"d2",dz);
 to_header("ic33","n3",niter,"o3",0.,"d3",1.);

 float *ggc11=new float[nnx*nnz]();
 float *ggc13=new float[nnx*nnz]();
 float *ggc33=new float[nnx*nnz]();

 for(int iter=1;iter<niter;++iter){
  fprintf(stderr,"iteration %d, obj func %f, grad2 %f\n",iter,f[iter-1],g[iter-1]);

  float pTg=dot_product(pc11,gc11,nnx*nnz)+dot_product(pc13,gc13,nnx*nnz)+dot_product(pc33,gc33,nnx*nnz);
  if(pTg>=0.){
   fprintf(stderr,"Search direction is not a decent direction. Take steepest decent.\n");
   mynegate(pc11,gc11,nnx*nnz);
   mynegate(pc13,gc13,nnx*nnz);
   mynegate(pc33,gc33,nnx*nnz);
  }

  float a=line_search(c11,c13,c33,f[iter],ggc11,ggc13,ggc33,f[iter-1],gc11,gc13,gc33,pc11,pc13,pc33,a0,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
  a0=a;
  fprintf(stderr,"step length %f\n",a);

  g[iter]=dot_product(ggc11,ggc11,nnx*nnz)+dot_product(ggc13,ggc13,nnx*nnz)+dot_product(ggc33,ggc33,nnx*nnz);
  
  float temp=dot_product(ggc11,gc11,nnx*nnz)+dot_product(ggc13,gc13,nnx*nnz)+dot_product(ggc33,gc33,nnx*nnz);

  float bPR=(g[iter]-temp)/g[iter-1];
  float bFR=g[iter]/g[iter-1];
  float b;
  if(bPR<-bFR) b=-bFR;
  else if(bPR>bFR) b=bFR;
  else b=bPR;
  fprintf(stderr,"b %f\n",b);

  lin_comb(pc11,-1.,ggc11,b,pc11,nnx*nnz); 
  lin_comb(pc13,-1.,ggc13,b,pc13,nnx*nnz); 
  lin_comb(pc33,-1.,ggc33,b,pc33,nnx*nnz); 

  float *pt;
  pt=gc11;gc11=ggc11;ggc11=pt;
  pt=gc13;gc13=ggc13;ggc13=pt;
  pt=gc33;gc33=ggc33;ggc33=pt;

  write("objfunc",f+iter,1,std::ios_base::app);
  write("normg2",g+iter,1,std::ios_base::app);
  
  write("ic11",c11,nnx*nnz,std::ios_base::app);
  write("ic13",c13,nnx*nnz,std::ios_base::app);
  write("ic33",c33,nnx*nnz,std::ios_base::app);
 }
 
 delete []gc11;delete []gc13;delete []gc33;
 delete []pc11;delete []pc13;delete []pc33;
 delete []ggc11;delete []ggc13;delete []ggc33;
 return;
}
