module pad_mod
 implicit none
 contains 
 
 subroutine pad2d(x,nx,nz,npad)
  real,intent(inout)::x(:,:)
  integer,intent(in)::nx,nz,npad
  integer::i
  real,parameter::pi=4.*atan(1.)
 !$omp sections
 !$omp section
 x(1:npad,1:npad)=x(npad+1,npad+1)
 !$omp section
 x(nx+npad+1:nx+2*npad,1:npad)=x(nx+npad,npad+1)
 !$omp section
 x(1:npad,nz+npad+1:nz+2*npad)=x(npad+1,nz+npad)
 !$omp section
 x(nx+npad+1:nx+2*npad,nz+npad+1:nz+2*npad)=x(nx+npad,nz+npad)
 !$omp end sections 
 
 !$omp parallel do default(shared) private(i)
 do i=1,npad
  x(i,npad+1:nz+npad)=x(npad+1,npad+1:nz+npad)
  x(nx+npad+i,npad+1:nz+npad)=x(nx+npad,npad+1:nz+npad)
  x(npad+1:nx+npad,i)=x(npad+1:nx+npad,npad+1)
  x(npad+1:nx+npad,nz+npad+i)=x(npad+1:nx+npad,nz+npad)
 enddo
 !$omp end parallel do
 end subroutine
 
 subroutine pad3d(x,nx,ny,nz,npad)
  real,intent(inout)::x(:,:,:)
  integer,intent(in)::nx,ny,nz,npad
  integer::i,j
  !$omp sections
  !$omp section
  x(1:npad,1:npad,1:npad)=x(npad+1,npad+1,npad+1)
  !$omp section
  x(nx+npad+1:nx+2*npad,1:npad,1:npad)=x(nx+npad,npad+1,npad+1)
  !$omp section
  x(1:npad+1,ny+npad+1:ny+2*npad,1:npad)=x(npad+1,ny+npad,npad+1)
  !$omp section
  x(nx+npad+1:nx+2*npad,ny+npad+1:ny+2*npad,1:npad)=x(nx+npad,ny+npad,npad+1)
  !$omp section
  x(1:npad,1:npad,nz+npad+1:nz+2*npad)=x(npad+1,npad+1,nz+npad)
  !$omp section
  x(nx+npad+1:nx+2*npad,1:npad,nz+npad+1:nz+2*npad)=x(nx+npad,npad+1,nz+npad)
  !$omp section
  x(1:npad+1,ny+npad+1:ny+2*npad,nz+npad+1:nz+2*npad)=x(npad+1,ny+npad,nz+npad)
  !$omp section
  x(nx+npad+1:nx+2*npad,ny+npad+1:ny+2*npad,nz+npad+1:nz+2*npad)=x(nx+npad,ny+npad,nz+npad)
  !$omp end sections
  
  !$omp parallel do default(shared) private(i,j)
  do j=1,npad
   do i=1,npad
    x(i,j,npad+1:nz+npad)=x(npad+1,npad+1,npad+1:nz+npad)
    x(nx+npad+i,j,npad+1:nz+npad)=x(nx+npad,npad+1,npad+1:nz+npad)
    x(i,ny+npad+j,npad+1:nz+npad)=x(npad+1,ny+npad,npad+1:nz+npad)
    x(nx+npad+i,ny+npad+j,npad+1:nz+npad)=x(nx+npad,ny+npad,npad+1:nz+npad)
    
    x(i,npad+1:ny+npad,j)=x(npad+1,npad+1:ny+npad,npad+1)
    x(nx+npad+i,npad+1:ny+npad,j)=x(nx+npad,npad+1:ny+npad,npad+1)
    x(i,npad+1:ny+npad,nz+npad+j)=x(npad+1,npad+1:ny+npad,nz+npad)
    x(nx+npad+i,npad+1:ny+npad,nz+npad+j)=x(nx+npad,npad+1:ny+npad,nz+npad)
    
    x(npad+1:nx+npad,i,j)=x(npad+1:nx+npad,npad+1,npad+1)
    x(npad+1:nx+npad,ny+npad+i,j)=x(npad+1:nx+npad,ny+npad,npad+1)
    x(npad+1:nx+npad,i,nz+npad+j)=x(npad+1:nx+npad,npad+1,nz+npad)
    x(npad+1:nx+npad,ny+npad+i,nz+npad+j)=x(npad+1:nx+npad,ny+npad,nz+npad)
   enddo
  enddo
  !$omp end parallel do
  
  !$omp parallel do default(shared) private(i)
  do i=1,npad
   x(i,npad+1:ny+npad,npad+1:nz+npad)=x(npad+1,npad+1:ny+npad,npad+1:nz+npad)
   x(nx+npad+i,npad+1:ny+npad,npad+1:nz+npad)=x(nx+npad,npad+1:ny+npad,npad+1:nz+npad)
   
   x(npad+1:nx+npad,i,npad+1:nz+npad)=x(npad+1:nx+npad,npad+1,npad+1:nz+npad)
   x(npad+1:nx+npad,ny+npad+i,npad+1:nz+npad)=x(npad+1:nx+npad,ny+npad,npad+1:nz+npad)
   
   x(npad+1:nx+npad,npad+1:ny+npad,i)=x(npad+1:nx+npad,npad+1:ny+npad,npad+1)
   x(npad+1:nx+npad,npad+1:ny+npad,nz+npad+i)=x(npad+1:nx+npad,npad+1:ny+npad,nz+npad)
  enddo
  !$omp end parallel do
 end subroutine 
 
end module

