#include <cmath>
#include <omp.h>
#include <cstdio>

#include "myio.h"
#include "init.h"
#include "wave.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f;
 get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate;
 get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d=new float[nr*nnt*ns]();
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();

 int opt;
 get_param("opt",opt);

 if(opt==0){
  init_model("c11",c11,nx,nz,npad); //c11=eps
  init_model("c13",c13,nx,nz,npad); //c13=del
  init_model("c33",c33,nx,nz,npad); //c33=v
 }
 else{
  float *v=new float[nnx*nnz]();
  init_model("v",v,nx,nz,npad); //c33=v
  float *eps=new float[nnx*nnz]();
  init_model("eps",eps,nx,nz,npad); //c11=eps
  float *del=new float[nnx*nnz]();
  init_model("del",del,nx,nz,npad); //c13=del
 
  float v0,eps0,del0;
  get_param("v0",v0);
  get_param("eps0",eps0);
  get_param("del0",del0);
 
  VEpsDel2Cij(c11,c13,c33,v,eps,del,v0,eps0,del0,nnx*nnz);
  delete []v;delete []eps;delete []del;
 }
 
 syntheticData(d,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);

 write("d",d,nr*nnt*ns);
 to_header("d","n1",nr,"o1",o_r,"d1",dr);
 to_header("d","n2",nnt,"o2",0.,"d2",rate);
 to_header("d","n3",ns,"o3",os,"d3",ds);

 delete []wavelet;
 delete []c11;
 delete []c13;
 delete []c33;
 delete []d;
 delete []sloc;
 delete []rloc;
 delete []taper;

 myio_close();
 return 0;
}
