#include <omp.h>
#include <cstdio>
#include <cstdlib>

#include "init.h"
#include "wave.h"
#include "kernels.h"

void syntheticData0(float *d,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 for(int is=0;is<ns;++is){
  int slocx=sloc[0+is*2];
  int slocz=sloc[1+is*2];
  syntheticDataShot(d+is*nr*nnt,c11,c13,c33,wavelet,slocx,slocz,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,ratio);
 }
 return;
}

void syntheticData1(float *d,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 printf("Total # GPUs = %d\n",nGPUs);
 
 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  printf("Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   int is=pass*nGPUs+i;
   int slocx=sloc[0+is*2];
   int slocz=sloc[1+is*2];
   cudaSetDevice(i);
   syntheticDataShot(d+is*nr*nnt,c11,c13,c33,wavelet,slocx,slocz,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,ratio);
  }
  shotLeft-=nGPUsNeed;
 }

 return;
}

void syntheticDataShot(float *d,const float *c11,const float *c13,const float *c33,const float *wavelet,int slocx,int slocz,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,int ratio){

 init_coeffs();

 int *d_rloc;
 cudaMalloc(&d_rloc,2*nr*sizeof(int));
 cudaMemcpy(d_rloc,rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
  
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;

 float *d_c11,*d_c13,*d_c33;
 cudaMalloc(&d_c11,nnx*nnz*sizeof(float));
 cudaMalloc(&d_c13,nnx*nnz*sizeof(float));
 cudaMalloc(&d_c33,nnx*nnz*sizeof(float));
 cudaMemcpy(d_c11,c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 cudaMemcpy(d_c13,c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 cudaMemcpy(d_c33,c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
 float *d_taper;
 cudaMalloc(&d_taper,nnx*nnz*sizeof(float));
 cudaMemcpy(d_taper,taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);

 float *d_d;
 cudaMalloc(&d_d,nr*sizeof(float));

 dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
 dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

 int slocxz=slocx+slocz*nnx;

 float *p0,*p1,*q0,*q1;
 cudaMalloc(&p0,nnx*nnz*sizeof(float)); cudaMemset(p0,0,nnx*nnz*sizeof(float));
 cudaMalloc(&p1,nnx*nnz*sizeof(float)); cudaMemset(p1,0,nnx*nnz*sizeof(float));
 cudaMalloc(&q0,nnx*nnz*sizeof(float)); cudaMemset(q0,0,nnx*nnz*sizeof(float));
 cudaMalloc(&q1,nnx*nnz*sizeof(float)); cudaMemset(q1,0,nnx*nnz*sizeof(float));
 
 inject<<<1,1>>>(p1,q1,wavelet[0],slocxz,dt2);

 abc<<<grid,block>>>(p1,q1,d_taper);
 
 if(ratio==1){
  record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d,p1,q1,d_rloc,nr,nnx);
  cudaMemcpy(d+nr,d_d,nr*sizeof(float),cudaMemcpyDeviceToHost);
 }
 
 for(int it=2;it<nt;++it){
  //printf("Time step it=%d\n",it);

  forwardCD<<<grid,block>>>(p0,q0,p1,q1,d_c11,d_c13,d_c33,dx2,dz2,dt2);

  inject<<<1,1>>>(p0,q0,wavelet[it-1],slocxz,dt2);

  abc<<<grid,block>>>(p0,q0,d_taper);
  abc<<<grid,block>>>(p1,q1,d_taper);
  
  if(it%ratio==0){
   record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d,p0,q0,d_rloc,nr,nnx);
   cudaMemcpy(d+(it/ratio)*nr,d_d,nr*sizeof(float),cudaMemcpyDeviceToHost);
  }

  float *pt=p0; 
  p0=p1;
  p1=pt;
  pt=q0;
  q0=q1;
  q1=pt;
 }

 cudaFree(d_c11);cudaFree(d_c13);cudaFree(d_c33);cudaFree(d_taper);
 cudaFree(p0);cudaFree(p1);cudaFree(q0);cudaFree(q1);cudaFree(d_d);

 return;
}

void syntheticData(float *d,const float *c11,const float *c13,const float *c33,const float *wavelet,const int *sloc,int ns,const int *rloc,int nr,const float *taper,int nx,int nz,int nt,int npad,float dx,float dz,float dt,float rate){
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 int nnx=nx+2*npad,nnz=nz+2*npad;
 float dx2=dx*dx,dz2=dz*dz,dt2=dt*dt;
 
 int nGPUs;
 cudaGetDeviceCount(&nGPUs);
 printf("Total # GPUs = %d\n",nGPUs);

 int **d_rloc=new int*[nGPUs]();
 float **d_c11=new float*[nGPUs]();
 float **d_c13=new float*[nGPUs]();
 float **d_c33=new float*[nGPUs]();
 float **d_taper=new float*[nGPUs]();
 float **d_d=new float*[nGPUs]();
 float **p0=new float*[nGPUs]();
 float **q0=new float*[nGPUs]();
 float **p1=new float*[nGPUs]();
 float **q1=new float*[nGPUs]();

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  
  init_coeffs();
 
  cudaMalloc(&d_rloc[i],2*nr*sizeof(int));
  cudaMemcpy(d_rloc[i],rloc,2*nr*sizeof(int),cudaMemcpyHostToDevice);
   
  cudaMalloc(&d_c11[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c13[i],nnx*nnz*sizeof(float));
  cudaMalloc(&d_c33[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_c11[i],c11,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c13[i],c13,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  cudaMemcpy(d_c33[i],c33,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
  
  cudaMalloc(&d_taper[i],nnx*nnz*sizeof(float));
  cudaMemcpy(d_taper[i],taper,nnx*nnz*sizeof(float),cudaMemcpyHostToDevice);
 
  cudaMalloc(&d_d[i],nr*sizeof(float));
 
  cudaMalloc(&p0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&p1[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q0[i],nnx*nnz*sizeof(float)); 
  cudaMalloc(&q1[i],nnx*nnz*sizeof(float)); 
 }
 
 int npasses=(ns+nGPUs-1)/nGPUs;
 int shotLeft=ns;

 for(int pass=0;pass<npasses;++pass){
  int nGPUsNeed=min(shotLeft,nGPUs);
  printf("Pass %d, # GPUs %d\n",pass,nGPUsNeed);
  #pragma omp parallel for num_threads(nGPUsNeed)
  for(int i=0;i<nGPUsNeed;++i){
   cudaSetDevice(i);

   int is=pass*nGPUs+i;
   int slocxz=sloc[0+is*2]+sloc[1+is*2]*nnx;

   dim3 block(BLOCK_DIM_X,BLOCK_DIM_Y);
   dim3 grid((nnx-2*RADIUS)/BLOCK_DIM_X,(nnz-2*RADIUS)/BLOCK_DIM_Y);

   cudaMemset(p0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q0[i],0,nnx*nnz*sizeof(float));
   cudaMemset(p1[i],0,nnx*nnz*sizeof(float));
   cudaMemset(q1[i],0,nnx*nnz*sizeof(float));

   inject<<<1,1>>>(p1[i],q1[i],wavelet[0],slocxz,dt2);
  
   abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
   
   if(ratio==1){
    record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p1[i],q1[i],d_rloc[i],nr,nnx);
    cudaMemcpy(d+nr+is*nr*nnt,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
   }
   
   for(int it=2;it<nt;++it){
    //printf("Time step it=%d\n",it);
  
    forwardCD<<<grid,block>>>(p0[i],q0[i],p1[i],q1[i],d_c11[i],d_c13[i],d_c33[i],dx2,dz2,dt2);
  
    inject<<<1,1>>>(p0[i],q0[i],wavelet[it-1],slocxz,dt2);
  
    abc<<<grid,block>>>(p0[i],q0[i],d_taper[i]);
    abc<<<grid,block>>>(p1[i],q1[i],d_taper[i]);
    
    if(it%ratio==0){
     record<<<(nr+BLOCK_DIM_X-1)/BLOCK_DIM_X,BLOCK_DIM_X>>>(d_d[i],p0[i],q0[i],d_rloc[i],nr,nnx);
     cudaMemcpy(d+(it/ratio)*nr+is*nr*nnt,d_d[i],nr*sizeof(float),cudaMemcpyDeviceToHost);
    }
  
    float *pt=p0[i]; 
    p0[i]=p1[i];
    p1[i]=pt;
    pt=q0[i];
    q0[i]=q1[i];
    q1[i]=pt;
   }
  }
  shotLeft-=nGPUsNeed;
 }

 for(int i=0;i<nGPUs;++i){
  cudaSetDevice(i);
  cudaFree(d_rloc[i]);
  cudaFree(d_c11[i]);cudaFree(d_c13[i]);cudaFree(d_c33[i]);
  cudaFree(d_taper[i]);
  cudaFree(d_d[i]);
  cudaFree(p0[i]);cudaFree(p1[i]);cudaFree(q0[i]);cudaFree(q1[i]);
 }
 
 delete []d_rloc;
 delete []d_c11;
 delete []d_c13;
 delete []d_c33;
 delete []d_taper;
 delete []d_d;
 delete []p0;
 delete []q0;
 delete []p1;
 delete []q1;

 return;
}
