#include <cmath>
#include <omp.h>
#include <cstdio>
#include <cstdlib>

#include "myio.h"
#include "init.h"
#include "wave.h"
#include "solvers.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f;
 get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);
 
 float *c33=new float[nnx*nnz]();
 init_model("c33",c33,nx,nz,npad); //c33=v
 float *c11=new float[nnx*nnz]();
 init_model("c11",c11,nx,nz,npad); //c11=eps
 float *c13=new float[nnx*nnz]();
 init_model("c13",c13,nx,nz,npad); //c13=del
 
// #pragma omp parallel for default(shared)
// for(int iz=0;iz<nnz;++iz){
//  for(int ix=0;ix<nnx;++ix){
//   int i=ix+iz*nnx; 
//   c33[i]=c33[i]*c33[i]; //c33=v*v
//   c11[i]=c33[i]*(1.+2.*c11[i]); //c11=v*v*(1+2*eps)
//   c13[i]=c33[i]*sqrt(1.+2.*c13[i]); //c13=v*v*sqrt(1+2*del)
//  }
// }
 
 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate;
 get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *gc11=new float[nnx*nnz](); read("gc11",gc11,nnx*nnz);
 float *gc13=new float[nnx*nnz](); read("gc13",gc13,nnx*nnz);
 float *gc33=new float[nnx*nnz](); read("gc33",gc33,nnx*nnz);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float *pc11=new float[nnx*nnz]();
 float *pc13=new float[nnx*nnz]();
 float *pc33=new float[nnx*nnz]();

 printf("Before CG\n"); 
// memcheck();

 GNcg(pc11,pc13,pc33,gc11,gc13,gc33,c11,c13,c33,d0,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);

// write("pc11",pc11,nnx*nnz);
// to_header("pc11","n1",nnx,"o1",-dx*npad,"d1",dx);
// to_header("pc11","n2",nnz,"o2",-dz*npad,"d2",dz);
// write("pc13",pc13,nnx*nnz*nnt);
// to_header("pc13","n1",nnx,"o1",-dx*npad,"d1",dx);
// to_header("pc13","n2",nnz,"o2",-dz*npad,"d2",dz);
// write("pc33",pc33,nnx*nnz);
// to_header("pc33","n1",nnx,"o1",-dx*npad,"d1",dx);
// to_header("pc33","n2",nnz,"o2",-dz*npad,"d2",dz);
 
 printf("Done writing results\n");

 delete []wavelet;
 delete []c11;delete []c13;delete []c33;
 delete []d0;
 delete []sloc;
 delete []rloc;
 delete []taper;
 delete []gc11;delete []gc13;delete []gc33;
 delete []pc11;delete []pc13;delete []pc33;

 printf("Done\n");
 
 myio_close();
 return 0;
}
