#include <cmath>
#include <omp.h>
#include <cstdio>

#include "myio.h"
#include "mylib.h"
#include "init.h"
#include "wave.h"

using namespace std;

int main(int argc,char **argv){
 myio_init(argc,argv);

int nnx,nnz,nx,nz,nt,npad;
 float ox,oz,ot,dx,dz,dt;
 init2d(nnx,nnz,nx,nz,nt,dx,dz,dt,ox,oz,ot,npad);

 float f;
 get_param("f",f);
 float *wavelet=new float[nt]();
 ricker(wavelet,f,nt,dt);

 float *c11=new float[nnx*nnz]();
 float *c13=new float[nnx*nnz]();
 float *c33=new float[nnx*nnz]();
 
 int padded;
 get_param("padded",padded);
 
 if(padded==0){
  init_model("c11",c11,nx,nz,npad); //c11=eps
  init_model("c13",c13,nx,nz,npad); //c13=del
  init_model("c33",c33,nx,nz,npad); //c33=v
 } 
 else{ 
  read("c11",c11,nnx*nnz);
  read("c13",c13,nnx*nnz);
  read("c33",c33,nnx*nnz);
 }

 int ns;
 float os,ds,zs;
 init_shot(ns,ds,os,zs);
 int *sloc=new int[2*ns]();
 init_shot_loc(sloc,ns,ds,os,zs,dx,dz,ox,oz,npad);
 
 int nr;
 float o_r,dr,zr;
 init_rec(nr,dr,o_r,zr);
 int *rloc=new int[2*nr]();
 init_rec_loc(rloc,nr,dr,o_r,zr,dx,dz,ox,oz,npad);
 
 float rate;
 get_param("rate",rate);
 int ratio=rate/dt+0.5;
 int nnt=(nt-1)/ratio+1;
 float *d0=new float[nr*nnt*ns]();
 read("data",d0,nr*nnt*ns);
 
 float *taper=new float[nnx*nnz]();
 init_abc(taper,nx,nz,npad);
 
 float *gc11=new float[nnx*nnz]();
 float *gc13=new float[nnx*nnz]();
 float *gc33=new float[nnx*nnz]();
 check_stability(c11,c13,c33,nnx,nnz,npad);
 gradient(gc11,gc13,gc33,d0,c11,c13,c33,wavelet,sloc,ns,rloc,nr,taper,nx,nz,nt,npad,dx,dz,dt,rate);
 
 write("gc11",gc11,nnx*nnz);
 to_header("gc11","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("gc11","n2",nnz,"o2",-dz*npad,"d2",dz);
 write("gc13",gc13,nnx*nnz);
 to_header("gc13","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("gc13","n2",nnz,"o2",-dz*npad,"d2",dz);
 write("gc33",gc33,nnx*nnz);
 to_header("gc33","n1",nnx,"o1",-dx*npad,"d1",dx);
 to_header("gc33","n2",nnz,"o2",-dz*npad,"d2",dz);
 
 delete []wavelet;delete []c11;delete []c13;delete []c33;delete []d0;delete []sloc;delete []rloc;delete []taper;delete []gc11;delete []gc13;delete []gc33;

 myio_close();
 return 0;
}
